//
// Public header compilation test program for the Fast Light Tool Kit (FLTK).
//
// Copyright 1998-2024 by Bill Spitzak and others.
//
// This library is free software. Distribution and use rights are outlined in
// the file "COPYING" which should have been included with this file.  If this
// file is missing or damaged, see the license at:
//
//     https://www.fltk.org/COPYING.php
//
// Please see the following page on how to report bugs and issues:
//
//     https://www.fltk.org/bugs.php
//

// The following header file is generated by CMake and includes all
// public FLTK headers:

#include "include_all.h"

#include <stdio.h>

int main() {
  printf("This program includes all FLTK header files for compilation.\n");
  printf("It should be compiled with `g++ (clang++) -Wshadow ...` or similar.\n");
  printf("This program does nothing.\n");
  return 0;
}
