
static const char * const comment_text[] = {
  // GNU Public License/GPL Header
    "//\n"
    "// NameOfProgram, ShortDescription\n"
    "// Copyright (C) YYYY  NameOfAuthor\n"
    "//\n"
    "// This program is free software: you can redistribute it and/or\n"
    "// modify it under the terms of the GNU General Public License\n"
    "// as published by the Free Software Foundation, either version 3\n"
    "// of the License, or (at your option) any later version.\n"
    "//\n"
    "// This program is distributed in the hope that it will be useful,\n"
    "// but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
    "// GNU General Public License for more details.\n"
    "//\n"
    "// You should have received a copy of the GNU General Public License\n"
    "// along with this program. If not, see <https://www.gnu.org/licenses/>.\n"
    "//\n",
  // GNU Public License/GPL Footer
    "\n//\n"
    "// NameOfProgram, ShortDescription\n"
    "// Copyright (C) YYYY  NameOfAuthor\n"
    "//",
  // GNU Public License/LGPL Header
    "//\n"
    "// NameOfLibrary, ShortDescription\n"
    "// Copyright (C) YYYY  NameOfAuthor\n"
    "//\n"
    "// This library is free software: you can redistribute it and/or\n"
    "// modify it under the terms of the GNU Lesser General Public License\n"
    "// as published by the Free Software Foundation, either version 3\n"
    "// of the License, or (at your option) any later version.\n"
    "//\n"
    "// This library is distributed in the hope that it will be useful,\n"
    "// but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    "// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n"
    "// GNU Lesser General Public License for more details.\n"
    "//\n"
    "// You should have received a copy of the GNU Lesser General Public License\n"
    "// along with this library. If not, see <https://www.gnu.org/licenses/>.\n"
    "//\n",
  // GNU Public License/LGPL Footer
    "\n//\n"
    "// NameOfLibrary, ShortDescription\n"
    "// Copyright (C) YYYY  NameOfAuthor\n"
    "//",
  // FLTK/Header
    "//\n"
    "// ... for the Fast Light Tool Kit (FLTK).\n"
    "//\n"
    "// Copyright 1998-2023 by Bill Spitzak and others.\n"
    "//\n"
    "// This library is free software. Distribution and use rights are outlined in\n"
    "// the file \"COPYING\" which should have been included with this file. If this\n"
    "// file is missing or damaged, see the license at:\n"
    "//\n"
    "//     https://www.fltk.org/COPYING.php\n"
    "//\n"
    "// Please see the following page on how to report bugs and issues:\n"
    "//\n"
    "//     https://www.fltk.org/bugs.php\n"
    "//\n"
  };
