#define _jcopy_block_row fltk__jcopy_block_row
#define _jcopy_sample_rows fltk__jcopy_sample_rows
#define _jdiv_round_up fltk__jdiv_round_up
#define _jinit_1pass_quantizer fltk__jinit_1pass_quantizer
#define _jinit_2pass_quantizer fltk__jinit_2pass_quantizer
#define _jinit_arith_decoder fltk__jinit_arith_decoder
#define _jinit_arith_encoder fltk__jinit_arith_encoder
#define _jinit_c_coef_controller fltk__jinit_c_coef_controller
#define _jinit_c_main_controller fltk__jinit_c_main_controller
#define _jinit_c_master_control fltk__jinit_c_master_control
#define _jinit_c_prep_controller fltk__jinit_c_prep_controller
#define _jinit_color_converter fltk__jinit_color_converter
#define _jinit_color_deconverter fltk__jinit_color_deconverter
#define _jinit_compress_master fltk__jinit_compress_master
#define _jinit_d_coef_controller fltk__jinit_d_coef_controller
#define _jinit_d_main_controller fltk__jinit_d_main_controller
#define _jinit_d_post_controller fltk__jinit_d_post_controller
#define _jinit_downsampler fltk__jinit_downsampler
#define _jinit_forward_dct fltk__jinit_forward_dct
#define _jinit_huff_decoder fltk__jinit_huff_decoder
#define _jinit_huff_encoder fltk__jinit_huff_encoder
#define _jinit_input_controller fltk__jinit_input_controller
#define _jinit_inverse_dct fltk__jinit_inverse_dct
#define _jinit_marker_reader fltk__jinit_marker_reader
#define _jinit_marker_writer fltk__jinit_marker_writer
#define _jinit_master_decompress fltk__jinit_master_decompress
#define _jinit_memory_mgr fltk__jinit_memory_mgr
#define _jinit_merged_upsampler fltk__jinit_merged_upsampler
#define _jinit_upsampler fltk__jinit_upsampler
#define _jpeg_CreateCompress fltk__jpeg_CreateCompress
#define _jpeg_CreateDecompress fltk__jpeg_CreateDecompress
#define _jpeg_abort fltk__jpeg_abort
#define _jpeg_abort_compress fltk__jpeg_abort_compress
#define _jpeg_abort_decompress fltk__jpeg_abort_decompress
#define _jpeg_add_quant_table fltk__jpeg_add_quant_table
#define _jpeg_alloc_huff_table fltk__jpeg_alloc_huff_table
#define _jpeg_alloc_quant_table fltk__jpeg_alloc_quant_table
#define _jpeg_aritab fltk__jpeg_aritab
#define _jpeg_calc_jpeg_dimensions fltk__jpeg_calc_jpeg_dimensions
#define _jpeg_calc_output_dimensions fltk__jpeg_calc_output_dimensions
#define _jpeg_consume_input fltk__jpeg_consume_input
#define _jpeg_copy_critical_parameters fltk__jpeg_copy_critical_parameters
#define _jpeg_core_output_dimensions fltk__jpeg_core_output_dimensions
#define _jpeg_default_colorspace fltk__jpeg_default_colorspace
#define _jpeg_default_qtables fltk__jpeg_default_qtables
#define _jpeg_destroy fltk__jpeg_destroy
#define _jpeg_destroy_compress fltk__jpeg_destroy_compress
#define _jpeg_destroy_decompress fltk__jpeg_destroy_decompress
#define _jpeg_fdct_10x10 fltk__jpeg_fdct_10x10
#define _jpeg_fdct_10x5 fltk__jpeg_fdct_10x5
#define _jpeg_fdct_11x11 fltk__jpeg_fdct_11x11
#define _jpeg_fdct_12x12 fltk__jpeg_fdct_12x12
#define _jpeg_fdct_12x6 fltk__jpeg_fdct_12x6
#define _jpeg_fdct_13x13 fltk__jpeg_fdct_13x13
#define _jpeg_fdct_14x14 fltk__jpeg_fdct_14x14
#define _jpeg_fdct_14x7 fltk__jpeg_fdct_14x7
#define _jpeg_fdct_15x15 fltk__jpeg_fdct_15x15
#define _jpeg_fdct_16x16 fltk__jpeg_fdct_16x16
#define _jpeg_fdct_16x8 fltk__jpeg_fdct_16x8
#define _jpeg_fdct_1x1 fltk__jpeg_fdct_1x1
#define _jpeg_fdct_1x2 fltk__jpeg_fdct_1x2
#define _jpeg_fdct_2x1 fltk__jpeg_fdct_2x1
#define _jpeg_fdct_2x2 fltk__jpeg_fdct_2x2
#define _jpeg_fdct_2x4 fltk__jpeg_fdct_2x4
#define _jpeg_fdct_3x3 fltk__jpeg_fdct_3x3
#define _jpeg_fdct_3x6 fltk__jpeg_fdct_3x6
#define _jpeg_fdct_4x2 fltk__jpeg_fdct_4x2
#define _jpeg_fdct_4x4 fltk__jpeg_fdct_4x4
#define _jpeg_fdct_4x8 fltk__jpeg_fdct_4x8
#define _jpeg_fdct_5x10 fltk__jpeg_fdct_5x10
#define _jpeg_fdct_5x5 fltk__jpeg_fdct_5x5
#define _jpeg_fdct_6x12 fltk__jpeg_fdct_6x12
#define _jpeg_fdct_6x3 fltk__jpeg_fdct_6x3
#define _jpeg_fdct_6x6 fltk__jpeg_fdct_6x6
#define _jpeg_fdct_7x14 fltk__jpeg_fdct_7x14
#define _jpeg_fdct_7x7 fltk__jpeg_fdct_7x7
#define _jpeg_fdct_8x16 fltk__jpeg_fdct_8x16
#define _jpeg_fdct_8x4 fltk__jpeg_fdct_8x4
#define _jpeg_fdct_9x9 fltk__jpeg_fdct_9x9
#define _jpeg_fdct_float fltk__jpeg_fdct_float
#define _jpeg_fdct_ifast fltk__jpeg_fdct_ifast
#define _jpeg_fdct_islow fltk__jpeg_fdct_islow
#define _jpeg_finish_compress fltk__jpeg_finish_compress
#define _jpeg_finish_decompress fltk__jpeg_finish_decompress
#define _jpeg_finish_output fltk__jpeg_finish_output
#define _jpeg_free_large fltk__jpeg_free_large
#define _jpeg_free_small fltk__jpeg_free_small
#define _jpeg_get_large fltk__jpeg_get_large
#define _jpeg_get_small fltk__jpeg_get_small
#define _jpeg_has_multiple_scans fltk__jpeg_has_multiple_scans
#define _jpeg_idct_10x10 fltk__jpeg_idct_10x10
#define _jpeg_idct_10x5 fltk__jpeg_idct_10x5
#define _jpeg_idct_11x11 fltk__jpeg_idct_11x11
#define _jpeg_idct_12x12 fltk__jpeg_idct_12x12
#define _jpeg_idct_12x6 fltk__jpeg_idct_12x6
#define _jpeg_idct_13x13 fltk__jpeg_idct_13x13
#define _jpeg_idct_14x14 fltk__jpeg_idct_14x14
#define _jpeg_idct_14x7 fltk__jpeg_idct_14x7
#define _jpeg_idct_15x15 fltk__jpeg_idct_15x15
#define _jpeg_idct_16x16 fltk__jpeg_idct_16x16
#define _jpeg_idct_16x8 fltk__jpeg_idct_16x8
#define _jpeg_idct_1x1 fltk__jpeg_idct_1x1
#define _jpeg_idct_1x2 fltk__jpeg_idct_1x2
#define _jpeg_idct_2x1 fltk__jpeg_idct_2x1
#define _jpeg_idct_2x2 fltk__jpeg_idct_2x2
#define _jpeg_idct_2x4 fltk__jpeg_idct_2x4
#define _jpeg_idct_3x3 fltk__jpeg_idct_3x3
#define _jpeg_idct_3x6 fltk__jpeg_idct_3x6
#define _jpeg_idct_4x2 fltk__jpeg_idct_4x2
#define _jpeg_idct_4x4 fltk__jpeg_idct_4x4
#define _jpeg_idct_4x8 fltk__jpeg_idct_4x8
#define _jpeg_idct_5x10 fltk__jpeg_idct_5x10
#define _jpeg_idct_5x5 fltk__jpeg_idct_5x5
#define _jpeg_idct_6x12 fltk__jpeg_idct_6x12
#define _jpeg_idct_6x3 fltk__jpeg_idct_6x3
#define _jpeg_idct_6x6 fltk__jpeg_idct_6x6
#define _jpeg_idct_7x14 fltk__jpeg_idct_7x14
#define _jpeg_idct_7x7 fltk__jpeg_idct_7x7
#define _jpeg_idct_8x16 fltk__jpeg_idct_8x16
#define _jpeg_idct_8x4 fltk__jpeg_idct_8x4
#define _jpeg_idct_9x9 fltk__jpeg_idct_9x9
#define _jpeg_idct_float fltk__jpeg_idct_float
#define _jpeg_idct_ifast fltk__jpeg_idct_ifast
#define _jpeg_idct_islow fltk__jpeg_idct_islow
#define _jpeg_input_complete fltk__jpeg_input_complete
#define _jpeg_mem_available fltk__jpeg_mem_available
#define _jpeg_mem_dest fltk__jpeg_mem_dest
#define _jpeg_mem_init fltk__jpeg_mem_init
#define _jpeg_mem_src fltk__jpeg_mem_src
#define _jpeg_mem_term fltk__jpeg_mem_term
#define _jpeg_natural_order fltk__jpeg_natural_order
#define _jpeg_natural_order2 fltk__jpeg_natural_order2
#define _jpeg_natural_order3 fltk__jpeg_natural_order3
#define _jpeg_natural_order4 fltk__jpeg_natural_order4
#define _jpeg_natural_order5 fltk__jpeg_natural_order5
#define _jpeg_natural_order6 fltk__jpeg_natural_order6
#define _jpeg_natural_order7 fltk__jpeg_natural_order7
#define _jpeg_new_colormap fltk__jpeg_new_colormap
#define _jpeg_open_backing_store fltk__jpeg_open_backing_store
#define _jpeg_quality_scaling fltk__jpeg_quality_scaling
#define _jpeg_read_coefficients fltk__jpeg_read_coefficients
#define _jpeg_read_header fltk__jpeg_read_header
#define _jpeg_read_raw_data fltk__jpeg_read_raw_data
#define _jpeg_read_scanlines fltk__jpeg_read_scanlines
#define _jpeg_resync_to_restart fltk__jpeg_resync_to_restart
#define _jpeg_save_markers fltk__jpeg_save_markers
#define _jpeg_set_colorspace fltk__jpeg_set_colorspace
#define _jpeg_set_defaults fltk__jpeg_set_defaults
#define _jpeg_set_linear_quality fltk__jpeg_set_linear_quality
#define _jpeg_set_marker_processor fltk__jpeg_set_marker_processor
#define _jpeg_set_quality fltk__jpeg_set_quality
#define _jpeg_simple_progression fltk__jpeg_simple_progression
#define _jpeg_start_compress fltk__jpeg_start_compress
#define _jpeg_start_decompress fltk__jpeg_start_decompress
#define _jpeg_start_output fltk__jpeg_start_output
#define _jpeg_std_error fltk__jpeg_std_error
#define _jpeg_std_huff_table fltk__jpeg_std_huff_table
#define _jpeg_std_message_table fltk__jpeg_std_message_table
#define _jpeg_stdio_dest fltk__jpeg_stdio_dest
#define _jpeg_stdio_src fltk__jpeg_stdio_src
#define _jpeg_suppress_tables fltk__jpeg_suppress_tables
#define _jpeg_write_coefficients fltk__jpeg_write_coefficients
#define _jpeg_write_m_byte fltk__jpeg_write_m_byte
#define _jpeg_write_m_header fltk__jpeg_write_m_header
#define _jpeg_write_marker fltk__jpeg_write_marker
#define _jpeg_write_raw_data fltk__jpeg_write_raw_data
#define _jpeg_write_scanlines fltk__jpeg_write_scanlines
#define _jpeg_write_tables fltk__jpeg_write_tables
#define _jround_up fltk__jround_up
